﻿using Microsoft.AspNetCore.Authentication;
using Microsoft.AspNetCore.Authorization;
using Microsoft.AspNetCore.Mvc;
using System;
using System.Collections.Generic;
using System.Net;
using System.Net.Http;
using System.Text;
using System.Threading.Tasks;

using Newtonsoft.Json;

namespace TestClient.Controllers
{
    /// <summary>
    /// 홈 컨트롤러
    /// </summary>
    public class HomeController : Controller
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// HTTP 클라이언트 팩토리
        /// </summary>
        private readonly IHttpClientFactory factory;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - HomeController(factory)

        /// <summary>
        /// 생성자
        /// </summary>
        /// <param name="factory">HTTP 클라이언트 팩토리</param>
        public HomeController(IHttpClientFactory factory)
        {
            this.factory = factory;
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 인덱스 페이치 처리하기 - Index()

        /// <summary>
        /// 인덱스 페이치 처리하기
        /// </summary>
        /// <returns>액션 결과</returns>
        public IActionResult Index()
        {
            return View();
        }

        #endregion
        #region 비밀 페이지 처리하기 - Secret()

        /// <summary>
        /// 비밀 페이지 처리하기
        /// </summary>
        /// <returns>액션 결과</returns>
        [Authorize]
        public async Task<IActionResult> Secret()
        {
            HttpResponseMessage authorizationServerResponse = await SecureRequest(() => GETSecureRequest("https://localhost:44300/secret/index"));

            if(authorizationServerResponse.IsSuccessStatusCode)
            {
                string result = await authorizationServerResponse.Content.ReadAsAsync<string>();

                ViewData["AuthorizationServerData"] = result;
            }

            HttpResponseMessage apiServerResponse = await SecureRequest(() => GETSecureRequest("https://localhost:44310/secret/index"));

            if(apiServerResponse.IsSuccessStatusCode)
            {
                string result = await apiServerResponse.Content.ReadAsAsync<string>();

                ViewData["APIServerData"] = result;
            }

            return View();
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region GET 보안 요청하기 - GETSecureRequest(url)

        /// <summary>
        /// GET 보안 요청하기
        /// </summary>
        /// <param name="url">URL</param>
        /// <returns>HTTP 응답 메시지</returns>
        private async Task<HttpResponseMessage> GETSecureRequest(string url)
        {
            string accessToken = await HttpContext.GetTokenAsync("access_token");

            HttpClient client = this.factory.CreateClient();

            client.DefaultRequestHeaders.Add("Authorization", $"Bearer {accessToken}");

            HttpResponseMessage response = await client.GetAsync(url);

            return response;
        }

        #endregion
        #region 액세스 토큰 갱신하기 - RefreshAccessToken()

        /// <summary>
        /// 액세스 토큰 갱신하기
        /// </summary>
        /// <returns>태스크</returns>
        private async Task RefreshAccessToken()
        {
            string refreshToken = await HttpContext.GetTokenAsync("refresh_token");

            HttpClient httpClient = this.factory.CreateClient();

            Dictionary<string, string> requestDictionary = new Dictionary<string, string>
            {
                ["grant_type"   ] = "refresh_token",
                ["refresh_token"] = refreshToken
            };

            string requestURL = "https://localhost:44300/oauth/token";

            HttpRequestMessage request = new HttpRequestMessage(HttpMethod.Post, requestURL)
            {
                Content = new FormUrlEncodedContent(requestDictionary)
            };

            string sourceAuthorization          = "username:password";
            byte[] sourceAuthorizationByteArray = Encoding.UTF8.GetBytes(sourceAuthorization);
            var    targetAuthorization          = Convert.ToBase64String(sourceAuthorizationByteArray);

            request.Headers.Add("Authorization", $"Basic {targetAuthorization}");

            HttpResponseMessage response = await httpClient.SendAsync(request);

            string responseJSON = await response.Content.ReadAsStringAsync();

            Dictionary<string, string> responseDictionary = JsonConvert.DeserializeObject<Dictionary<string, string>>(responseJSON);

            string newAccessToken  = responseDictionary.GetValueOrDefault("access_token" );
            string newRefreshToken = responseDictionary.GetValueOrDefault("refresh_token");

            AuthenticateResult authenticateResult = await HttpContext.AuthenticateAsync("TestClientCookie");

            authenticateResult.Properties.UpdateTokenValue("access_token" , newAccessToken );
            authenticateResult.Properties.UpdateTokenValue("refresh_token", newRefreshToken);

            await HttpContext.SignInAsync("TestClientCookie", authenticateResult.Principal, authenticateResult.Properties);
        }

        #endregion
        #region 보안 요청하기 - SecureRequest(requestFunction)

        /// <summary>
        /// 보안 요청하기
        /// </summary>
        /// <param name="requestFunction">요청 함수</param>
        /// <returns>HTTP 응답 메시지 태스크</returns>
        private async Task<HttpResponseMessage> SecureRequest(Func<Task<HttpResponseMessage>> requestFunction)
        {
            HttpResponseMessage response = await requestFunction();

            if(response.StatusCode == HttpStatusCode.Unauthorized)
            {
                await RefreshAccessToken();

                response = await requestFunction();
            }

            return response;
        }

        #endregion
    }
}